"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the updater of the PyFunceble configuration.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020, 2021, 2022, 2023, 2024, 2025 Dead Hosts Contributors
    Copyright (c) 2019, 2020. 2021, 2022, 2023, 2024 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import importlib.resources
import logging
import os
from typing import Optional

from PyFunceble.helpers.dict import DictHelper
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper
from PyFunceble.helpers.file import FileHelper
from PyFunceble.helpers.merge import Merge

import dead_hosts.launcher.defaults.links
import dead_hosts.launcher.defaults.paths
import dead_hosts.launcher.defaults.pyfunceble
from dead_hosts.launcher.info_manager import InfoManager
from dead_hosts.launcher.updater.base import UpdaterBase


class PyFuncebleConfigUpdater(UpdaterBase):
    """
    Provides the updated of the PyFunceble configuration.
    """

    def __init__(self, info_manager: InfoManager) -> None:
        self.pyfunceble_config_file_instance = FileHelper(
            os.path.join(
                info_manager.pyfunceble_config_dir, ".PyFunceble.overwrite.yaml"
            )
        )

        super().__init__(info_manager)

    @property
    def authorized(self) -> bool:
        return (
            self.info_manager.platform_optout is True
            and not self.info_manager.own_management
        ) or EnvironmentVariableHelper("PLATFORM_WORKER").get_value() is not None

    @staticmethod
    def get_commit_message(message: str, ping: Optional[str] = None) -> str:
        """
        Provides the commit message to use.
        """

        if ping:
            return f"{message} | cc {ping} | "
        return message

    def pre(self) -> "PyFuncebleConfigUpdater":
        logging.info(
            "Started to update %r.",
            self.pyfunceble_config_file_instance.path,
        )

        return self

    def post(self) -> "PyFuncebleConfigUpdater":
        logging.info(
            "Finished to update %r",
            self.pyfunceble_config_file_instance.path,
        )

        return self

    def start(self) -> "PyFuncebleConfigUpdater":
        with importlib.resources.path(
            "PyFunceble.data.infrastructure", ".PyFunceble_production.yaml"
        ) as file_path:
            local_version = DictHelper(
                DictHelper().from_yaml_file(str(file_path))
            ).flatten()

        local_version = Merge(
            dead_hosts.launcher.defaults.pyfunceble.CONFIGURATION
        ).into(local_version, strict=True)

        if self.info_manager.custom_pyfunceble_config and isinstance(
            self.info_manager.custom_pyfunceble_config, dict
        ):
            logging.info(
                "Custom PyFunceble configuration given, "
                "appending them to the local configuration file."
            )

            local_version = Merge(self.info_manager.custom_pyfunceble_config).into(
                local_version, strict=True
            )

        if self.info_manager.ping:
            logging.info("Ping names given, appending them to the commit message.")

            local_version["cli_testing.ci.end_commit_message"] = (
                self.get_commit_message(
                    local_version["cli_testing.ci.end_commit_message"],
                    ping=self.info_manager.get_ping_for_commit(),
                )
            )

        if not self.info_manager.platform_optout:
            local_version = Merge(
                dead_hosts.launcher.defaults.pyfunceble.PLATFORM_PERSISTENT_CONFIG
            ).into(local_version, strict=True)
        else:
            local_version = Merge(
                dead_hosts.launcher.defaults.pyfunceble.PERSISTENT_CONFIG
            ).into(local_version, strict=True)

        if FileHelper(
            os.path.join(
                self.info_manager.WORKSPACE_DIR,
                dead_hosts.launcher.defaults.paths.EXAMPLE_INFO_FILENAME,
            )
        ).exists():
            local_version["cli_testing.ci.active"] = False
            # Default behavior of PyFunceble since 4.0.0b12.
            local_version["cli_testing.autocontinue"] = False

        local_version = DictHelper(local_version).unflatten()

        DictHelper(local_version).to_yaml_file(
            self.pyfunceble_config_file_instance.path
        )

        logging.debug("Configuration:\n%s", self.pyfunceble_config_file_instance.read())
        logging.info(
            "Configuration file %r updated.",
            self.pyfunceble_config_file_instance.path,
        )

        return self



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the orchestration logic. In other words, put everything
together to process a test.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020, 2021, 2022, 2023, 2024, 2025 Dead Hosts Contributors
    Copyright (c) 2019, 2020. 2021, 2022, 2023, 2024 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import logging
import os
import sys
import tempfile
from datetime import datetime, timezone
from typing import Optional

import PyFunceble
import PyFunceble.facility
import requests
from PyFunceble.cli.continuous_integration.exceptions import (
    ContinuousIntegrationException,
    StopExecution,
)
from PyFunceble.cli.continuous_integration.github_actions import GitHubActions
from PyFunceble.helpers.download import DownloadHelper
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper
from PyFunceble.helpers.file import FileHelper

import dead_hosts.launcher.defaults.envs
import dead_hosts.launcher.defaults.paths
from dead_hosts.launcher.authorization import Authorization
from dead_hosts.launcher.command import Command
from dead_hosts.launcher.info_manager import InfoManager
from dead_hosts.launcher.platform import PlatformOrchestration
from dead_hosts.launcher.updater.all import execute_all_updater
from dead_hosts.launcher.updater.pyfunceble_config import PyFuncebleConfigUpdater


class Orchestration:
    """
    Orchester the test launch.
    """

    info_manager: Optional[InfoManager] = None
    authorization_handler: Optional[Authorization] = None

    origin_file: Optional[FileHelper] = None
    output_file: Optional[FileHelper] = None

    def __init__(
        self,
        save: bool = False,
        end: bool = False,
        authorize: bool = False,
        update: bool = False,
    ) -> None:
        self.info_manager = InfoManager()

        git_name = EnvironmentVariableHelper("GIT_NAME")
        git_email = EnvironmentVariableHelper("GIT_EMAIL")

        if git_email.exists() and "funilrys" in git_email.get_value():
            git_name.set_value(dead_hosts.launcher.defaults.envs.GIT_NAME)
            git_email.set_value(dead_hosts.launcher.defaults.envs.GIT_EMAIL)

        EnvironmentVariableHelper("PYFUNCEBLE_OUTPUT_LOCATION").set_value(
            self.info_manager.WORKSPACE_DIR
            if self.info_manager.platform_optout is True
            else os.path.join(tempfile.gettempdir(), "pyfunceble", "output")
        )

        EnvironmentVariableHelper("PYFUNCEBLE_CONFIG_DIR").set_value(
            self.info_manager.pyfunceble_config_dir
        )

        self.authorization_handler = Authorization(self.info_manager)

        self.origin_file = FileHelper(
            os.path.join(
                self.info_manager.WORKSPACE_DIR,
                dead_hosts.launcher.defaults.paths.ORIGIN_FILENAME,
            )
        )

        self.output_file = FileHelper(
            os.path.join(
                self.info_manager.WORKSPACE_DIR,
                dead_hosts.launcher.defaults.paths.OUTPUT_FILENAME,
            )
        )

        if not authorize:
            logging.info("Origin file: %r", self.origin_file.path)
            logging.info("Output file: %r", self.output_file.path)

        if authorize:
            self.run_authorize()
        elif save:
            self.run_autosave()
        elif update:
            execute_all_updater(self.info_manager)
        elif end:
            self.run_end()
        else:
            logging.info("Checking authorization to run.")

            if EnvironmentVariableHelper("PLATFORM_WORKER").get_value():
                self.run_platform_worker()
            elif self.authorization_handler.is_platform_authorized():
                self.run_platform_sync()
            elif self.authorization_handler.is_test_authorized():
                PyFunceble.facility.ConfigLoader.start()
                self.fetch_file_to_test()

                self.run_test()
            else:
                logging.info(
                    "Not authorized to run a test until %r (current time) > %r",
                    datetime.now(),
                    self.authorization_handler.next_authorization_time,
                )
                sys.exit(0)

    def fetch_file_to_test(self) -> "Orchestration":
        """
        Provides the latest version of the file to test.
        """

        if self.authorization_handler.is_refresh_authorized():
            logging.info("We are authorized to refresh the lists! Let's do that.")
            logging.info("Raw Link: %r", self.info_manager.raw_link)

            if self.info_manager.raw_link:
                DownloadHelper(self.info_manager.raw_link).download_text(
                    destination=self.origin_file.path
                )

                logging.info(
                    "Could get the new version of the list. Updating the download time."
                )

                self.info_manager["last_download_datetime"] = datetime.now(timezone.utc)

                try:
                    self.info_manager["last_download_timestamp"] = self.info_manager[
                        "last_download_datetime"
                    ].timestamp()
                except (OSError, ValueError, OverflowError):
                    self.info_manager["last_download_timestamp"] = 0.0
            elif self.origin_file.exists():
                logging.info(
                    "Raw link not given or is empty. Let's work with %r.",
                    self.origin_file.path,
                )

                self.origin_file.read()

                logging.info("Emptying the download time.")

                self.info_manager["last_download_datetime"] = datetime.fromtimestamp(0)
                try:
                    self.info_manager["last_download_timestamp"] = self.info_manager[
                        "last_download_datetime"
                    ].timestamp()
                except (OSError, ValueError, OverflowError):
                    self.info_manager["last_download_timestamp"] = 0.0
            else:
                logging.info(
                    "Could not find %s. Generating empty content to test.",
                    self.origin_file.path,
                )

                self.origin_file.write("# No content yet.", overwrite=True)

                logging.info("Emptying the download time.")

                self.info_manager["last_download_datetime"] = datetime.fromtimestamp(0)

                try:
                    self.info_manager["last_download_timestamp"] = self.info_manager[
                        "last_download_datetime"
                    ].timestamp()
                except (OSError, ValueError, OverflowError):
                    self.info_manager["last_download_timestamp"] = 0.0

            logging.info("Updated %r.", self.origin_file.path)

        return self

    def run_platform_worker(self):
        """
        Run a test of the input list.
        """

        # Ensure that everything is up-to-date.
        execute_all_updater(self.info_manager)

        self.write_trigger()

        logging.info("Starting PyFunceble %r ...", PyFunceble.__version__)

        EnvironmentVariableHelper("PYFUNCEBLE_BYPASS_BYPASS").set_value("True")

        Command("pyfunceble platform").run_to_stdout()

        if not dead_hosts.launcher.defaults.envs.GITHUB_TOKEN:
            self.run_end()

    def run_test(self):
        """
        Run a test of the input list.
        """

        if not self.info_manager.currently_under_test:
            self.info_manager["currently_under_test"] = True

            self.info_manager["start_datetime"] = datetime.now(timezone.utc)

            try:
                self.info_manager["start_timestamp"] = self.info_manager[
                    "start_datetime"
                ].timestamp()
            except (OSError, ValueError, OverflowError):
                self.info_manager["start_timestamp"] = 0.0

            self.info_manager["finish_datetime"] = datetime.fromtimestamp(0)

            try:
                self.info_manager["finish_timestamp"] = self.info_manager[
                    "finish_datetime"
                ].timestamp()
            except (OSError, ValueError, OverflowError):
                self.info_manager["finish_timestamp"] = 0.0

        self.info_manager["latest_part_start_datetime"] = datetime.now(timezone.utc)

        try:
            self.info_manager["latest_part_start_timestamp"] = self.info_manager[
                "latest_part_start_datetime"
            ].timestamp()
        except (OSError, ValueError, OverflowError):
            self.info_manager["latest_part_start_timestamp"] = 0.0

        self.info_manager["latest_part_finish_datetime"] = datetime.fromtimestamp(0)

        try:
            self.info_manager["latest_part_finish_timestamp"] = self.info_manager[
                "latest_part_finish_datetime"
            ].timestamp()
        except (OSError, ValueError, OverflowError):
            self.info_manager["latest_part_finish_timestamp"] = 0.0

        logging.info("Updated all timestamps.")
        logging.info("Starting PyFunceble %r ...", PyFunceble.__version__)

        Command(f"pyfunceble -f {self.origin_file.path}").run_to_stdout()

        self.write_trigger()
        if not dead_hosts.launcher.defaults.envs.GITHUB_TOKEN:
            self.run_end()

    def run_autosave(self):
        """
        Run the autosave logic of the administration file.

        .. warning::
            This is just about the administration file not PyFunceble.
        """

        self.info_manager["latest_part_finish_datetime"] = datetime.now(timezone.utc)
        try:
            self.info_manager["latest_part_finish_timestamp"] = self.info_manager[
                "latest_part_finish_datetime"
            ].timestamp()
        except (OSError, ValueError, OverflowError):
            self.info_manager["latest_part_finish_timestamp"] = 0.0

        self.write_trigger()

        logging.info("Updated all timestamps.")

    def run_end(self):
        """
        Run the end logic.
        """

        self.info_manager["currently_under_test"] = False

        self.info_manager["latest_part_finish_datetime"] = datetime.now(timezone.utc)

        try:
            self.info_manager["latest_part_finish_timestamp"] = self.info_manager[
                "latest_part_finish_datetime"
            ].timestamp()
        except (OSError, ValueError, OverflowError):
            self.info_manager["latest_part_finish_timestamp"] = 0.0

        self.info_manager["finish_datetime"] = self.info_manager[
            "latest_part_finish_datetime"
        ]
        try:
            self.info_manager["finish_timestamp"] = self.info_manager[
                "finish_datetime"
            ].timestamp()
        except (OSError, ValueError, OverflowError):
            self.info_manager["finish_timestamp"] = 0.0

        logging.info("Updated all timestamps and indexes that needed to be updated.")

        pyfunceble_active_list = FileHelper(
            os.path.join(
                self.info_manager.WORKSPACE_DIR,
                "output",
                dead_hosts.launcher.defaults.paths.ORIGIN_FILENAME,
                "domains",
                "ACTIVE",
                "list",
            )
        )

        logging.info("PyFunceble ACTIVE list output: %s", pyfunceble_active_list.path)

        if pyfunceble_active_list.exists():
            logging.info(
                "%s exists, getting and formatting its content.",
                pyfunceble_active_list.path,
            )

            self.output_file.write(
                "\n".join(dead_hosts.launcher.defaults.paths.OUTPUT_FILE_HEADER)
                + "\n\n",
                overwrite=True,
            )

            with pyfunceble_active_list.open("r", encoding="utf-8") as file_stream:
                for line in file_stream:
                    if line.startswith("#"):
                        continue

                    self.output_file.write(line)

            self.output_file.write("\n")

            logging.info("Updated of the content of %r", self.output_file.path)

        self.write_trigger()

    def _fetch_workflows(self, token, repo):
        headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {token}",
        }

        if not token:
            del headers["Authorization"]

        req = requests.get(
            url=f"https://api.github.com/repos/{repo}/actions/workflows",
            headers=headers,
            timeout=30.0,
        )

        req.raise_for_status()

        result = {}

        for data in req.json()["workflows"]:
            result[data["id"]] = data

        return result

    def _is_job_running(self, token, repo):
        workflows = self._fetch_workflows(token, repo)
        ignore_status = [
            "completed",
            "cancelled",
            "failure",
            "neutral",
            "skipped",
            "success",
            "timed_out",
        ]
        headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {token}",
        }

        if not token:
            del headers["Authorization"]

        req = requests.get(
            url=f"https://api.github.com/repos/{repo}/actions/runs",
            headers=headers,
            timeout=30.0,
        )

        req.raise_for_status()

        response = req.json()["workflow_runs"]

        for data in response:
            if data["status"] in ignore_status:
                continue

            workflow_name = workflows[data["workflow_id"]]["name"].lower()

            if "checker" in workflow_name or "auth" in workflow_name:
                continue

            return True

        return False

    def write_trigger(self):
        """
        Write the trigger file.
        """

        with open(
            os.path.join(self.info_manager.WORKSPACE_DIR, ".trigger"),
            "w",
            encoding="utf-8",
        ) as file_stream:
            logging.info("Writing into: %s", file_stream.name)
            file_stream.write(str(datetime.now(timezone.utc).timestamp()) + "\n")

    def run_authorize(self):
        """
        Authorized and proceed the scheduling.
        """

        if not dead_hosts.launcher.defaults.envs.GITHUB_TOKEN:
            logging.critical("Cannot authorize: Token not found.")
            return None

        if self._is_job_running(
            token=dead_hosts.launcher.defaults.envs.GITHUB_TOKEN,
            repo=self.info_manager.repo,
        ):
            logging.critical("Cannot authorize: Job running.")
            return None

        try:
            ci_engine = GitHubActions(
                commit_message="[Dead-Hosts::Infrastructure][AuthChecker]"
            )
            ci_engine.init()
        except ContinuousIntegrationException:
            pass

        execute_all_updater(self.info_manager)

        self.write_trigger()

        try:
            ci_engine.apply_commit()
        except (StopExecution, ContinuousIntegrationException):
            pass

        logging.info("Successfully authorized.")

        return True

    def run_platform_sync(self):
        """
        Run the synchronization / Download against the platform.
        """

        logging.info("Launching the platform synchronization.")

        commit_message = PyFuncebleConfigUpdater.get_commit_message(
            f"[Final/Result][Dead-Hosts::"
            f"{dead_hosts.launcher.defaults.paths.GIT_BASE_NAME}]",
            ping=self.info_manager.get_ping_for_commit(),
        )

        try:
            ci_engine = GitHubActions(
                commit_message=commit_message, end_commit_message=commit_message
            )
            ci_engine.init()
        except ContinuousIntegrationException:
            pass

        try:
            # The commit is one of the last one.
            ci_engine.bypass()
        except (StopExecution, ContinuousIntegrationException):
            logging.info("No need to synchronize.")
            sys.exit(0)

        execute_all_updater(self.info_manager)

        platform = PlatformOrchestration(self.info_manager)

        # We upload the local information to the platform.
        platform.upload()

        try:
            # Download what's available - yet
            platform.download()
            self.write_trigger()
            ci_engine.apply_end_commit()
        except (StopExecution, ContinuousIntegrationException):
            pass

        logging.info("Successfully synchronized.")

        return True



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides our default paths.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020, 2021, 2022, 2023, 2024, 2025 Dead Hosts Contributors
    Copyright (c) 2019, 2020. 2021, 2022, 2023, 2024 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import os
import urllib.parse
from datetime import datetime, timezone
from typing import List

from dead_hosts.launcher.command import Command

CURRENT_DIRECTORY: str = os.getcwd()

GHA_WORKFLOW_DIR: str = os.path.join(".github", "workflows")
PYFUNCEBLE_NAMESPACE: str = ".pyfunceble"
PYFUNCEBLE_CONFIG_DIRECTORY: str = os.path.join(CURRENT_DIRECTORY, PYFUNCEBLE_NAMESPACE)

GHA_WORKFLOW_LOCAL_HASHES_FILENAME: str = "hs.json"
INFO_FILENAME: str = "info.json"
ORIGIN_FILENAME: str = "origin.list"
INPUT_FILENAME: str = "domains.list"
OUTPUT_FILENAME: str = "clean.list"
README_FILENAME: str = "README.md"

EXAMPLE_INFO_FILENAME: str = "info.example.json"

REMOTE_URL: str = Command("git remote get-url origin").execute().strip()
PARSED_REMOTE_URL: str = urllib.parse.urlparse(REMOTE_URL)

if not PARSED_REMOTE_URL.netloc:
    REPO_BASE = PARSED_REMOTE_URL.path.split(":", 1)[-1]
else:
    REPO_BASE = PARSED_REMOTE_URL.path[1:]

if REPO_BASE.endswith(".git"):
    REPO_BASE = REPO_BASE[:-4]

GIT_BASE_NAME: str = REPO_BASE.split("/", 1)[-1]
GIT_REPO_OWNER: str = REPO_BASE.split("/", 1)[0]

OUTPUT_FILE_HEADER: List[str] = [
    "# File generated by the Dead-Hosts project with the help of PyFunceble.",
    "# Dead-Hosts: https://github.com/dead-hosts",
    "# PyFunceble: https://pyfunceble.github.io",
    f"# Generation Time: {datetime.now(timezone.utc).isoformat()}",
]



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the updater of our pyfunceble configuration location.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020, 2021, 2022, 2023, 2024, 2025 Dead Hosts Contributors
    Copyright (c) 2019, 2020. 2021, 2022, 2023, 2024 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import logging
import os

from PyFunceble.helpers.directory import DirectoryHelper
from PyFunceble.helpers.file import FileHelper

from dead_hosts.launcher.updater.base import UpdaterBase


class PyFuncebleConfigLocationUpdater(UpdaterBase):
    """
    Provides the updater of our pyfunceble configuration location.
    """

    FILES_TO_MOVE: str = [
        ".pyfunceble_intern_downtime.json",
        ".PyFunceble_production.yaml",
        ".PyFunceble.yaml",
        "dir_structure.json",
        "hashes_tracker.json",
        "iana-domains-db.json",
        "inactive_db.json",
        "public-suffix.json",
        "user_agents.json",
        "whois_db.json",
    ]

    INACTIVE_FILES_TO_DELETE: str = ["inactive_db.json", "inactive.csv"]

    @property
    def authorized(self) -> bool:
        if self.info_manager.platform_optout is False:
            # Controller don't want to opt-out from the platform testing.
            # That means that we won't test anything locally.
            return False

        if not DirectoryHelper(self.info_manager.pyfunceble_config_dir).exists():
            return True

        file_helper = FileHelper()
        for file in self.FILES_TO_MOVE:
            if file_helper.set_path(
                os.path.join(self.info_manager.pyfunceble_config_dir, file)
            ).exists():
                return True

        if all(
            file_helper.set_path(
                os.path.join(self.info_manager.pyfunceble_config_dir, x)
            ).exists()
            for x in self.INACTIVE_FILES_TO_DELETE
        ):
            return True

        return False

    def pre(self) -> "PyFuncebleConfigLocationUpdater":
        logging.info(
            "Started maintenance of %r.",
            self.info_manager.pyfunceble_config_dir,
        )

        DirectoryHelper(self.info_manager.pyfunceble_config_dir).create()

        return self

    def post(self) -> "PyFuncebleConfigLocationUpdater":
        logging.info(
            "Finished maintenance %r",
            self.info_manager.pyfunceble_config_dir,
        )

        return self

    def start(self) -> "PyFuncebleConfigLocationUpdater":
        for file in self.FILES_TO_MOVE:
            source_file = FileHelper(
                os.path.join(self.info_manager.WORKSPACE_DIR, file)
            )
            destination_file = FileHelper(
                os.path.join(self.info_manager.pyfunceble_config_dir, file)
            )

            if source_file.exists():
                logging.info(
                    "Starting to move %r into %r.",
                    source_file.path,
                    destination_file.path,
                )

                destination_file.delete()
                source_file.move(destination_file.path)

                logging.info(
                    "Finished to move %r into %r",
                    source_file.path,
                    destination_file.path,
                )
            else:
                logging.info(
                    "Did not moved move %r into %r: It does not exists.",
                    source_file.path,
                    destination_file.path,
                )

            if (
                all(
                    FileHelper(
                        os.path.join(self.info_manager.pyfunceble_config_dir, x)
                    ).exists()
                    for x in self.INACTIVE_FILES_TO_DELETE
                )
                or FileHelper(
                    os.path.join(
                        self.info_manager.pyfunceble_config_dir,
                        self.INACTIVE_FILES_TO_DELETE[0],
                    )
                ).exists()
            ):
                logging.info(
                    "Starting to delete inactive files.",
                )

                for inactive_file in self.INACTIVE_FILES_TO_DELETE:
                    FileHelper(
                        os.path.join(
                            self.info_manager.pyfunceble_config_dir, inactive_file
                        )
                    ).delete()

                logging.info(
                    "Finished to delete inactive files.",
                )

        return self



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the base of all our updaters.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020, 2021, 2022, 2023, 2024, 2025 Dead Hosts Contributors
    Copyright (c) 2019, 2020. 2021, 2022, 2023, 2024 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import secrets
from typing import Optional

from dead_hosts.launcher.info_manager import InfoManager


class UpdaterBase:
    """
    Provides the base of all updates.

    :param info_manager:
        The info manager to work with.

    :raise TypeError:
        When the given :code:`info_manager` is not a real info manager.
    """

    do_not_start: bool = False
    info_manager: Optional[InfoManager] = None

    def __init__(self, info_manager: InfoManager) -> None:
        if not isinstance(info_manager, InfoManager):
            raise TypeError(info_manager)

        self.info_manager = info_manager

        if not self.do_not_start:
            self.start_after_authorization()

    @property
    def authorized(self) -> bool:
        """
        Provides the authorization to process.
        """

        return False

    def pre(self) -> "UpdaterBase":
        """
        Called before :code:`start`.
        """

        raise NotImplementedError()

    def post(self) -> "UpdaterBase":
        """
        Called after :code:`start`.
        """

        raise NotImplementedError()

    def start(self) -> "UpdaterBase":
        """
        Starts the update.
        """

        raise NotImplementedError()

    def start_after_authorization(self) -> "UpdaterBase":
        """
        Starts after checking the authorization.
        """

        if self.authorized:
            self.pre()
            self.start()
            self.post()

        return self

    @staticmethod
    def randomize_cron(given: str = None, randomize_index=(0,)) -> str:
        """
        Provides a random cron expression.
        """

        index2ranges = {
            0: (0, 59 + 1),
            1: (0, 23 + 1),
            2: (1, 31 + 1),
            3: (1, 12 + 1),
            4: (0, 6 + 1),
        }

        if given:
            dataset = given.split(" ")

            for index in randomize_index:
                first_part = str(secrets.choice(range(*index2ranges[index])))
                second_part = str(secrets.choice(range(*index2ranges[index])))

                if index not in randomize_index:
                    dataset[index] = f"{first_part}/{second_part}"
                else:
                    dataset[index] = f"{first_part}"
        else:
            dataset = ["*", "*", "*", "*", "*"]

            for index in randomize_index:
                dataset[index] = str(secrets.choice(range(*index2ranges[index])))

        return " ".join(dataset)



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the platform interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020, 2021, 2022, 2023, 2024, 2025 Dead Hosts Contributors
    Copyright (c) 2019, 2020. 2021, 2022, 2023, 2024 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import json
import logging
import os
import tempfile
from typing import Optional

import requests
from PyFunceble.helpers.environment_variable import EnvironmentVariableHelper
from PyFunceble.helpers.file import FileHelper

import dead_hosts.launcher.defaults.paths
from dead_hosts.launcher.info_manager import InfoManager


class PlatformOrchestration:
    """
    Provides the platform interface.
    """

    DEFAULT_API_URL = "https://api.example.com"
    DEFAULT_SCOPE = "example-username"

    _api_url = DEFAULT_API_URL
    _scope = DEFAULT_SCOPE
    __cached_header: dict = {}

    info_manager: InfoManager
    headers: dict = {}
    session: requests.Session

    def __init__(self, info_manager: InfoManager) -> None:
        self.info_manager = info_manager

        self.api_url = EnvironmentVariableHelper("DEV_PLATFORM_API_URL").get_value(
            default=None
        ) or EnvironmentVariableHelper("PLATFORM_API_URL").get_value(
            default=self.DEFAULT_API_URL
        )

        api_token = EnvironmentVariableHelper("DEV_PLATFORM_API_TOKEN").get_value(
            default=None
        ) or EnvironmentVariableHelper("PLATFORM_API_TOKEN").get_value(default=None)

        self.scope = EnvironmentVariableHelper("DEV_PLATFORM_SCOPE").get_value(
            default=None
        ) or EnvironmentVariableHelper("PLATFORM_SCOPE").get_value(
            default=self.DEFAULT_SCOPE
        )

        self.headers = {
            "Authorization": f"Bearer {api_token}",
            "Content-Type": "application/json",
        }

        self.session = requests.Session()
        self.session.headers.update(self.headers)

    @property
    def api_url(self) -> str:
        """
        Provides the API URL to use.
        """

        return self._api_url

    @api_url.setter
    def api_url(self, value: str) -> None:
        """
        Sets the API URL to use.
        """

        self._api_url = value.rstrip("/")

    @property
    def scope(self) -> str:
        """
        Provides the scope to use.
        """

        return self._scope

    @scope.setter
    def scope(self, value: str) -> None:
        """
        Sets the scope to use.
        """

        self._scope = value

    @property
    def container_id(self) -> str:
        """
        Provides the container ID to use.
        """

        return self.info_manager.platform_container_id

    @container_id.setter
    def container_id(self, value: str) -> None:
        """
        Sets the container ID to use.
        """

        self.info_manager["platform_container_id"] = value

    @property
    def container_canonical_name(self) -> str:
        """
        Provides the container name to use.
        """

        return self.info_manager.platform_container_name

    @container_canonical_name.setter
    def container_canonical_name(self, value: str) -> None:
        """
        Sets the container name to use.
        """

        self.info_manager["platform_container_name"] = value

    @property
    def container_description(self) -> str:
        """
        Provides the container description to use.
        """

        return self.info_manager.platform_description

    @container_description.setter
    def container_description(self, value: str) -> None:
        """
        Sets the container description to use.
        """

        self.info_manager["platform_description"] = value

    @property
    def remote_source_id(self) -> str:
        """
        Provides the remote source ID to use.
        """

        return self.info_manager.platform_remote_source_id

    @remote_source_id.setter
    def remote_source_id(self, value: str) -> None:
        """
        Sets the remote source ID to use.
        """

        self.info_manager["platform_remote_source_id"] = value

    @property
    def container_display_name(self) -> str:
        """
        Provides the container display name to use.
        """

        return self.info_manager.name

    @property
    def container_slug(self) -> str:
        """
        Provides the container slug to use.
        """

        return f"{self.scope}/{self.info_manager.platform_container_name}"

    @property
    def container_canonical_data(self) -> dict:
        """
        Provides the canonical data of the container.
        """

        return {
            "availability_test": True,
            "reputation_test": False,
            "syntax_test": True,
            "name": self.container_canonical_name,
            "display_name": self.container_canonical_name,
            "description": self.container_description,
            "public": False,
            "subject_type": "domain",
        }

    @property
    def remote_source_canonical_data(self) -> dict:
        """
        Provides the canonical data of the remote source.
        """

        if self.info_manager.raw_link:
            url = self.info_manager.raw_link
        else:
            url = (
                "https://raw.githubusercontent.com"
                f"/{self.info_manager.repo}/master/origin.list"
            )

        input_format = "plain"

        if self.info_manager.custom_pyfunceble_config:
            if (
                "adblock" in self.info_manager.custom_pyfunceble_config
                and self.info_manager.custom_pyfunceble_config["adblock"]
            ):
                input_format = "adblock"
            elif (
                "cli_decoding.adblock" in self.info_manager.custom_pyfunceble_config
                and self.info_manager.custom_pyfunceble_config["cli_decoding.adblock"]
            ):
                input_format = "adblock"

        return {
            "description": self.container_description,
            "url": url,
            "subject_type": "domain",
            "input_format": input_format,
            "link_type": "associated",
        }

    def search_matching_container(
        self, display_name: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[dict]:
        """
        Queries the matching container.
        """

        logging.info(
            "Searching for container matching name (%s) or display name (%s).",
            self.container_display_name if not display_name else display_name,
            self.container_canonical_name if not name else name,
        )

        search_info = {
            "approximate": False,
            "display_name": (
                self.container_display_name if not display_name else display_name
            ),
            "name": (self.container_canonical_name if not name else name),
            "username": self.scope,
        }

        params = {"limit": -1}

        try:
            req = self.session.post(
                f"{self.api_url}/v1/containers/search", params=params, json=search_info
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return None

        if req.status_code == 200:
            response = req.json()

            logging.debug(
                "Query-Matching-Container | URL: %s | Status Code: %s | "
                "Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            logging.info(
                "Found %d matching containers.",
                len(response),
            )

            return response

        logging.debug(
            "Query-Container | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info("Found no matching container.")

        return None

    def read_container(self):
        """
        Reads the container.
        """

        logging.info("Reading container matching name (%s).", self.container_slug)

        try:
            req: requests.Response = self.session.get(
                f"{self.api_url}/v1/containers/{self.container_slug}"
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return None

        if req.status_code == 200:
            response = req.json()

            logging.debug(
                "Read-Container | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            logging.info("Found container matching name (%s).", self.container_slug)

            return response

        logging.debug(
            "Read-Container | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info("Found no container matching name (%s).", self.container_slug)

        return None

    def read_remote_sources(self):
        """
        Reads the remote sources of the container.
        """

        logging.info(
            "Reading remote sources of container matching name (%s).",
            self.container_slug,
        )

        params = {"limit": -1}

        try:
            req = self.session.get(
                f"{self.api_url}/v1/containers/{self.container_slug}/remote-sources",
                params=params,
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return []

        if req.status_code == 200:
            response = req.json()

            logging.debug(
                "Read-Remote-Sources | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            logging.info(
                "Found %d remote sources of container matching name (%s).",
                len(response),
                self.container_slug,
            )

            return response

        if req.status_code == 204:
            logging.debug(
                "Read-Remote-Sources | URL: %s | Status Code: %s | NO CONTENT",
                req.url,
                req.status_code,
            )

            logging.info(
                "Found no remote sources of container  matching name (%s).",
                self.container_slug,
            )

            return []

        logging.debug(
            "Read-Remote-Sources | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Found no remote sources of container matching name (%s).",
            self.container_slug,
        )

        return []

    def read_export_rules(self):
        """
        Reads the export rules of the container.
        """

        logging.info(
            "Reading export rules of container matching name (%s).",
            self.container_slug,
        )

        params = {"limit": -1}

        try:
            req = self.session.get(
                f"{self.api_url}/v1/containers/{self.container_slug}/export-rules",
                params=params,
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return []

        if req.status_code == 200:
            response = req.json()

            logging.debug(
                "Read-Export-Rules | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            logging.info(
                "Found %d export rules of container matching name (%s).",
                len(response),
                self.container_slug,
            )

            return response

        if req.status_code == 204:
            logging.debug(
                "Read-Export-Rules | URL: %s | Status Code: %s | NO CONTENT",
                req.url,
                req.status_code,
            )

            logging.info(
                "Found no export rules of container matching name (%s).",
                self.container_slug,
            )

            return []

        logging.debug(
            "Read-Export-Rules | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Found no export rules of container matching name (%s).",
            self.container_slug,
        )

        return []

    def download_export_rules(self, export_rule_id: str, destination: str):
        """
        Downloads the given export rule.

        :param export_rule_id:
            The ID of the export rule to download.

        :param destination:
            The destination of the downloaded file.
        """

        logging.info(
            "Downloading export rule with ID: %s to %s",
            export_rule_id,
            destination,
        )

        try:
            req = self.session.get(
                f"{self.api_url}/v1/containers/{self.container_slug}"
                f"/export-rules/{export_rule_id}/plain",
                stream=True,
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return False

        if req.status_code == 200:
            with open(destination, "wb") as file_handler:
                for chunk in req.iter_content(chunk_size=1024):
                    file_handler.write(chunk)

            logging.info(
                "Successfully downloaded export rule with ID: %s to %s",
                export_rule_id,
                destination,
            )

            return True

        if req.status_code == 204:
            logging.info(
                "Export rule (%s) is empty.",
                export_rule_id,
            )

            return False

        logging.critical(
            "Download-Export-Rule | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        return False

    def create_container(self):
        """
        Creates the container.
        """

        logging.info("Creating container.")

        try:
            req = self.session.post(
                f"{self.api_url}/v1/containers", json=self.container_canonical_data
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return ""

        if req.status_code in (200, 202, 303):
            response = req.json()

            self.container_id = response["id"]
            self.container_canonical_name = response["name"]

            logging.info(
                "Successfully created container with ID: %s",
                self.info_manager.platform_container_id,
            )
            logging.debug(
                "Create-Container | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            return response

        logging.critical(
            "Create-Container | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info("Failed to create container.")

        try:
            return req.json()
        except json.JSONDecodeError:
            return req.text

    def create_remote_source(self):
        """
        Creates the remote source.

        .. warning::
            This method assumes that only 1 remote source is desired.
        """

        logging.info(
            "Creating remote source for container matching name (%s)",
            self.container_slug,
        )

        try:
            req = self.session.post(
                f"{self.api_url}/v1/containers/{self.container_slug}/remote-sources",
                json=self.remote_source_canonical_data,
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return ""

        if req.status_code in (200, 202, 303):
            response = req.json()

            self.remote_source_id = response["id"]

            logging.info(
                "Successfully created remote source with ID: %s",
                response["id"],
            )

            logging.debug(
                "Create-Remote-Source | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            return response

        logging.critical(
            "Create-Remote-Source | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Failed to create remote source for container matching name (%s)",
            self.container_slug,
        )

        try:
            return req.json()
        except json.JSONDecodeError:
            return req.text

    def update_container(self):
        """
        Updates the container.
        """

        logging.info(
            "Updating container matching name (%s)",
            self.container_slug,
        )

        try:
            req = self.session.patch(
                f"{self.api_url}/v1/containers/{self.container_slug}",
                json=self.container_canonical_data,
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return ""

        if req.status_code in (200, 202, 303):
            response = req.json()

            self.container_id = response["id"]
            self.container_canonical_name = response["name"]

            logging.info(
                "Successfully updated container matching name (%s)",
                self.container_slug,
            )

            logging.debug(
                "Update-Container | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            return req.json()

        logging.critical(
            "Update-Container | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Failed to update container matching name (%s)",
            self.container_slug,
        )

        try:
            return req.json()
        except json.JSONDecodeError:
            return req.text

    def update_remote_source(self):
        """
        Updates the remote source.
        """

        logging.info(
            "Updating remote source with ID (%s) for container matching name (%s)",
            self.remote_source_id,
            self.container_slug,
        )

        try:
            req = self.session.patch(
                f"{self.api_url}/v1/containers/{self.container_slug}"
                f"/remote-sources/{self.remote_source_id}",
                json=self.remote_source_canonical_data,
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return ""

        if req.status_code in (200, 202, 303):
            response = req.json()

            logging.info(
                "Successfully updated remote source with ID: %s",
                response["id"],
            )

            logging.debug(
                "Update-Remote-Source | URL: %s | Status Code: %s | Response:\n    %s",
                req.url,
                req.status_code,
                response,
            )

            return response

        logging.critical(
            "Update-Remote-Source | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Failed to update remote source with ID (%s) for container matching "
            "name (%s)",
            self.remote_source_id,
            self.container_slug,
        )

        try:
            return req.json()
        except json.JSONDecodeError:
            return req.text

    def delete_container(self):
        """
        Deletes the container.

        :param container_id:
            The ID of the container to delete.

            .. note::
                If not given, we will follow the platform_container_id index.
        :type container_id: uuid.UUID
        """

        logging.info(
            "Deleting container matching name (%s).",
            self.container_slug,
        )

        try:
            req = self.session.delete(
                f"{self.api_url}/v1/containers/{self.container_slug}"
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return False

        if req.status_code == 204:
            logging.info(
                "Successfully deleted container matching name (%s).",
                self.container_slug,
            )

            self.container_id = None
            self.remote_source_id = None

            return True

        logging.critical(
            "Delete-Container | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Failed to delete container matching name (%s).",
            self.container_slug,
        )

        return False

    def delete_remote_source(self):
        """
        Deletes the remote source.
        """

        logging.info(
            "Deleting remote source with ID (%s) for container matching name (%s)",
            self.remote_source_id,
            self.container_slug,
        )

        try:
            req = self.session.delete(
                f"{self.api_url}/v1/containers/{self.container_slug}"
                f"/remote-sources/{self.remote_source_id}"
            )
        except requests.exceptions.RequestException:
            logging.critical("Failed to communicate with platform. Opt-Out forgotten?")
            return False

        if req.status_code == 204:
            logging.info(
                "Successfully deleted remote source with ID: %s",
                self.remote_source_id,
            )

            self.remote_source_id = None

            return True

        logging.critical(
            "Delete-Remote-Source | URL: %s | Status Code: %s | Response:\n    %s",
            req.url,
            req.status_code,
            req.text,
        )

        logging.info(
            "Failed to delete remote source with ID (%s) for container matching "
            "name (%s)",
            self.remote_source_id,
            self.container_slug,
        )

        return False

    def upload(self):
        """
        Ensures that all locally defined information are uploaded to the platform
        - for testing.
        """

        if not self.container_id:
            found_containers = self.search_matching_container()

            if found_containers:
                self.container_id = found_containers[0]["id"]
                self.container_canonical_name = found_containers[0]["name"]

                self.update_container()
            else:
                self.create_container()
        elif self.read_container():
            self.update_container()
        else:
            self.create_container()

        if not self.remote_source_id:
            found_remote_sources = self.read_remote_sources()

            if found_remote_sources:
                self.remote_source_id = found_remote_sources[0]["id"]
                self.update_remote_source()
            else:
                self.create_remote_source()
        else:
            dataset = self.read_remote_sources()

            if dataset:
                self.remote_source_id = dataset[0]["id"]

                self.update_remote_source()
            else:
                self.create_remote_source()

    def download(self):
        """
        Downloads the latest compiled version of the dataset.
        """

        if not self.container_id:
            found_containers = self.search_matching_container()

            if found_containers:
                self.container_id = found_containers[0]["id"]
                self.container_canonical_name = found_containers[0]["name"]

        if not self.container_id:
            logging.info("No container found. Aborting.")
            return False

        wanted_export_rules = []

        expected_availability_status = {
            "active": True,
            "inactive": False,
            "invalid": False,
        }
        expected_reputation_status = {
            "sane": False,
            "malicious": False,
            "invalid": False,
        }
        expected_syntax_status = {
            "valid": False,
            "invalid": False,
        }

        for export_rule in self.read_export_rules():
            if (
                all(
                    y == export_rule["availability"][x]
                    for x, y in expected_availability_status.items()
                )
                and all(
                    y == export_rule["reputation"][x]
                    for x, y in expected_reputation_status.items()
                )
                and all(
                    y == export_rule["syntax"][x]
                    for x, y in expected_syntax_status.items()
                )
            ):
                wanted_export_rules.append(export_rule)

        files = []

        for export_rule in wanted_export_rules:
            logging.debug("Downloading export rule: %r", export_rule)

            with tempfile.NamedTemporaryFile(
                delete=False, suffix="_export_rule"
            ) as tmp_filestream:
                if not self.download_export_rules(
                    export_rule["id"], tmp_filestream.name
                ):
                    continue

                tmp_filestream.seek(0)
                files.append(tmp_filestream.name)

        logging.info("Starting to merge the downloaded files.")

        # pylint: disable=consider-using-with
        temp_output_file = tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix="_output"
        )

        output_file = FileHelper(
            os.path.join(
                self.info_manager.WORKSPACE_DIR,
                dead_hosts.launcher.defaults.paths.OUTPUT_FILENAME,
            )
        )

        temp_output_file.write(
            "\n".join(dead_hosts.launcher.defaults.paths.OUTPUT_FILE_HEADER) + "\n\n"
        )

        written = False

        for file in files:
            file = FileHelper(file)

            with file.open("r", encoding="utf-8") as file_stream:
                for line in file_stream:
                    if line.startswith("#"):
                        continue

                    temp_output_file.write(line)
                    written = True

        temp_output_file.seek(0)

        logging.info("Finished to merge the downloaded files.")

        if written:
            logging.info("Moving the merged file to its final destination.")
            FileHelper(temp_output_file.name).move(output_file.path)
            logging.info("Moved the merged file to its final destination.")

        logging.info("Cleaning up temporary files.")
        for file in files:
            os.remove(file)
        logging.info("Cleaned up temporary files.")

        return True
